%%%%%%%%%%
% SeWeHGrius deduction API core RDF(S) predicates ("rules")
%
% Author: Arttu Valo
% CVS Id $Id: sewehgrius_rdfs_core_rules.pl,v 1.7 2004/02/28 00:50:00 kimvilja Exp $
%


%%%%%%%%%
% Generic constants for use in all predicates
%

rdfs_ns('http://www.w3.org/TR/1999/PR-rdf-schema-19990303#').
rdfs_ns('http://www.w3.org/2000/01/rdf-schema#').

rdfs_uri(ID,URI) :-
    rdfs_ns(NS),
    string_concat( NS, ID, URIS ),
    string_to_atom( URIS, URI ).

rdfs_property(ID,URI) :-
    rdfs_ns(NS),
    URI=NS:ID.

rdf_ns('http://www.w3.org/1999/02/22-rdf-syntax-ns#type').

rdf_uri(ID,URI) :-
    rdf_ns(NS),
    string_concat( NS, ID, URIS ),
    string_to_atom( URIS, URI ).

rdf_property(ID,URI) :-
    rdf_ns(NS),
    URI=NS:ID.

rdfs_class( C ) :-
    rdfs_normal_class( C ).

rdfs_class( C ) :-
    rdfs_metaclass( C ).

%rdfs_metaclass( C ) :-
%    rdfs_normal_class( C ).

rdfs_metaclass( C ) :-
    rdf_type( C, MC ), 
    rdfs_uri( 'Class', ClassURI ),
    rdfs_transitive_subClassOf( MC, ClassURI ),
    rdfs_normal_class( MC ).

rdfs_normal_class( C ) :-
    rdfs_uri( 'Class', ClassURI ),
    rdf_type( C, ClassURI ).

% Class is a class in rdf(s) ontology.
% Name is returned. It is the name of the class.
%rdfs_class_name( +Class, -Name )
rdfs_class_name( Class, Name ) :-
%    nonvar( Class ),
%    atomic( Class ),
    rdfs_property(label, LabelURI),
    rdf( Class, LabelURI, literal( Name ) ).

rdfs_class_name( Class, Name ) :-
%    nonvar( Class ),
    Class = Namespace:ID,
    string_concat( Namespace, ID, String3 ),
    string_to_atom( String3, ClassAtom ),
    rdfs_property(label, labelURI),
    rdf( ClassAtom, labelURI, literal( Name ) ).


%rdfs_label( +URI, -Label )
rdfs_label( URI, Label ) :-
%    nonvar( URI ),
    rdfs_class_name( URI, Label ).


%%%%%%%%%%
% RDF utilities
%


%rdf_type( ?URI, ?TypeURI )
rdf_type( URI, TypeURI ) :-
    rdf( URI, rdf:type, TypeURI ).


%%%%%%%%%%
% RDF Schema basic class hierarchy deduction rules
%


%rdfs_instanceOf( +InstanceURI, +ClassURI )
rdfs_instanceOf( InstanceURI, ClassURI ) :-
    rdf_uri('type',TypeURI),
    rdf( InstanceURI, TypeURI, ClassURI ).
%av: below: commented out, we want only direct connections
%rdfs_instanceOf( InstanceURI, ClassURI ) :-
%    rdf( InstanceURI, 'http://www.w3.org/1999/02/22-rdf-syntax-ns#type', IntermedClass ),
%    subClassOf( IntermedClass, ClassURI ).


%rdfs_class( +SubjectURI, -ClassURI )
rdfs_class( SubjectURI, ClassURI ) :-
    % AV: instanceoffing is done implicitly, so it should be redundant
    rdfs_uri('Class', RdfsClassURI),
    rdfs_instanceOf( SubjectURI, RdfsClassURI ),
    rdf( SubjectURI, RdfsClassURI, ClassURI ). % What does this mean????? (Kim doesn't understand...)
rdfs_class( SubjectURI, ClassURI ) :-
    % AV: instanceoffing is done implicitly, so it should be redundant
    rdfs_uri('Class', RdfsClassURI),
    rdfs_instanceOf( SubjectURI, RdfsClassURI ),
    rdf( SubjectURI, RdfsClassURI, ClassURI ).
%    rdf( Class, rdf:type, 'http://www.w3.org/TR/1999/PR-rdf-schema-19990303#Class' ).
%    rdf( SubjectURI, '???#instanceOf', Class ).
%AV: FIXXX: hasClass?!


%rdfs_superClassOf( ?SuperClass, ?SubClass )
rdfs_superClassOf( SuperClass, SubClass ) :-
    rdfs_property(subClassOf, SubClassOfProperty),
    rdf( SubClass, SubClassOfProperty, SuperClass ).

% AV: FIXXX: intermediate classes for super and subclasses as well??? NOPE, I'm saying now!

%rdfs_subClassOf( ?SubClass, ?SuperClass ) :-
rdfs_subClassOf( SubClass, SuperClass ) :-
    rdfs_superClassOf( SuperClass, SubClass ).

% for testing purposes
%rdfs_transitive_subClassOf( SubClass, SuperClass ) :-
%    format( 'rdfs_transitive_subClassOf( ~w, ~w )~n', [SubClass, SuperClass] ),
%    fail.

rdfs_transitive_subClassOf( SubClass, SuperClass, cache_or_direct ) :-
    rdfs_subClassOf( SubClass, SuperClass );
    get_from_cache(cache_rdfs_transitive_subClassOf( SubClass, SuperClass )).

rdfs_transitive_subClassOf( SubClass, SuperClass ) :-
    rdfs_subClassOf( SubClass, SuperClass ).

rdfs_transitive_subClassOf( SubClass, SuperClass ) :-
    get_from_cache(cache_rdfs_transitive_subClassOf( SubClass, SuperClass )).

rdfs_transitive_subClassOf( SubClass, SuperClass ) :-
    nonvar( SubClass ),
    rdfs_subClassOf( SubClass, SuperClass1 ),
    rdfs_transitive_subClassOf( SuperClass1, SuperClass ),
    add_to_cache(cache_rdfs_transitive_subClassOf( SubClass, SuperClass )).

rdfs_transitive_subClassOf( SubClass, SuperClass ) :-
    nonvar( SuperClass ),
    rdfs_subClassOf( SubClass1, SuperClass ),
    rdfs_transitive_subClassOf( SubClass, SubClass1 ),
    add_to_cache(cache_rdfs_transitive_subClassOf( SubClass, SuperClass )).

add_to_cache( Term ) :-
    catch( 
    (
        (
	    Term;
            assert(Term)
%            write('assert ')
        )
    ),
    Exception,
    (
        Exception=error(existence_error(procedure, _),_)->
	assert(Term);
%	write('assert ');
	true
%	format('Assert exception ~w~n', Exception)
    )
    ),
    !.

get_from_cache( Term ) :-
    catch( 
	(
	    Term
%	    write('get from cache ')
	),
        _,
        fail
    ).

cache_subclass_hierarchy :-
    rdfs_uri('Class', RdfsClassURI),
    forall( true, rdfs_transitive_subClassOf( S, RdfsClassURI ) ),
    rdfs_uri('Resource', RdfsResourceURI),
    forall( true, rdfs_transitive_subClassOf( S, RdfsResourceURI) ).

%rdfs_domainOf( +SubjectURI, -Domain )
rdfs_domainOf( SubjectURI, Domain ) :-
    rdfs_uri('domain', RdfsDomainURI),
    rdf( SubjectURI, RdfsDomainURI, Domain ).
    % uri(Domain). % no literals!?

%rdfs_rangeOf( +SubjectURI, -Range )
rdfs_rangeOf( SubjectURI, Range ) :-
    rdfs_uri('range', RdfsRangeURI),
    rdf( SubjectURI, RdfsRangeURI, Range ).
    % uri(Domain). % no literals!?


%%%%%%%%%%
% SeWeHGRius RDF(S) relationary predicates
%

%sewehgrius_relation_same_class( ?From, ?To )
sewehgrius_relation_same_class( From, To ) :-
    rdf( From, rdf:type, Class ),
    rdf( To, rdf:type, Class ),
    not( From = To ).

% From and To have same Property which point to the same instance.
%sewehgrius_relation_same_property( ?From, ?To, ?Property, ?PropertyURI, ?TypeA, ?TypeB )
sewehgrius_relation_same_property( From, To, Property, PropertyURI, TypeA, TypeB ) :-
    ignore( rdfs_class( From, TypeA ) ),  % they might not be classes!
    ignore( rdfs_class( To, TypeB ) ),    % they might not be classes!
    rdf( From, Property, PropertyURI ),
    not( Property = rdf:type ),
    rdf( To, Property, PropertyURI ),
    not( From = To ).
%    rdf( PropertyURI, rdf:type, Class ).
