#!/usr/bin/perl -w

##########
# Builds prolog's categories responses into the cache
# CVS Id $Id: cache_categories.pl,v 1.5 2003/12/18 15:21:05 kimvilja Exp $
# Author: Arttu Valo, spoiled by Kim Viljanen
#

use strict;

if ($#ARGV <= 2) {
    die( "Wrong amount of arguments...\nUSAGE: $0 [ontodella_home] [prolog] [category_dir]" );
}


my %CMDARGS;

for(my $i=0;$i<=$#ARGV; $i++) {
   if ($ARGV[$i] =~/^([^=]*?)=(.*?)$/) {
      my $pkey = $1;
      my $param = $2;
      $param =~ s/\s+$//;
      $CMDARGS{lc($pkey)}=$param;
   }
}

my $MODE_FORCE=0;

if ($CMDARGS{cparams} =~ /-f$/ || $CMDARGS{cparams} =~ /-f\s/) {
   $MODE_FORCE=1;
}

#foreach my $k (sort keys %CMDARGS) {
#   print "$k='".$CMDARGS{$k}."'\n";
#}

##########
# variables and handling command line arguments

my $ONTODELLA_HOME = $CMDARGS{home};
my $PROLOG_PARAM = $CMDARGS{plparams};
my $PROLOG_BINARY = $CMDARGS{pl}." $PROLOG_PARAM";
my $ONTODELLA_PARAMS = $CMDARGS{odparams};

my $ONTODELLA_CONFIG=$ONTODELLA_HOME."/conf/fms/ontodella_config.pl";
my $PROLOG_SOURCE = $ONTODELLA_HOME."/src/prolog/ontodella_server.pl";

my $CATEGORY_DIR = $CMDARGS{catdir};
my $CATEGORY_FILE = $CATEGORY_DIR . "/categories.rdf" ;
my $TEMP_FILE = $CATEGORY_DIR . "/cacher_cat_temp.tmp";

my $SEWEHGRIUS_SERVICE = "get_categories_";


print "PROLOG_BINARY: $PROLOG_BINARY \n";


if (! -d $CATEGORY_DIR ) 
{
    mkdir $CATEGORY_DIR;
    if (! -d $CATEGORY_DIR ) {
        die( "Couldn't create directory $CATEGORY_DIR\n" );
    }
    print "Added directory $CATEGORY_DIR\n";
}

if (-e $TEMP_FILE ) {
    if ($MODE_FORCE == 1) {
        unlink($TEMP_FILE);
    } else {
        die( "Temporary file $TEMP_FILE exists. Another $0 running?\n\nRemove the file before trying again or use the switch -f.\n" );
    }
}


##########
# Logic
#

my $cmd = $PROLOG_BINARY . " -s " . $PROLOG_SOURCE
    . " -g \"server,ontodella_request_handler:data_minimizer,ontodella_request_handler:test_categories,halt\" -- home=$ONTODELLA_HOME config=$ONTODELLA_CONFIG ". $ONTODELLA_PARAMS
    . " > $TEMP_FILE";

print "cmd=$cmd\n";
system( $cmd );


# Read the temp file back in and process it

my $temp_mode = 0;

open ( TEMPFILE, $TEMP_FILE ) or
    die "$0: failed to read from file; $!\n";
open ( OUTFILE, ">$CATEGORY_FILE" ) or
    die "$0: failed to write to file; $!\n";

while ( my $line = <TEMPFILE> )
{
    if ( $line =~ m/^<\?xml/ )
    {
        $temp_mode = 1;
    }
    elsif ( $line =~ m|^</rdf:RDF>| )
    {
        $line =~ s|^(</rdf:RDF>).*?$|$1|i; # remove text outside of the xml region.
        print OUTFILE $line;
        $temp_mode = 0;
    }

    if ( $temp_mode == 1 )
    {
        print OUTFILE $line;
    }
}
close( OUTFILE );
close( TEMPFILE );

unlink( $TEMP_FILE );