#!/usr/bin/perl -w

##########
# Builds prolog's relation responses into the cache
# CVS Id $Id: cache_relations.pl,v 1.2 2004/03/18 13:32:21 kimvilja Exp $
# Author: Arttu Valo
#

use strict;

##########
# Configuration
#

my $PROLOG_BINARY = "/home/tkt_fms/tools/prolog/bin/pl -G14M -L10M";
my $PROLOG_SOURCE = "ontodella_server.pl";
my $CACHE_DIR = "/tmp/tkt_fms/odcache/";
my $CACHE_FILE_SUFFIX = ".tmp";
my $TEMP_FILE_NAME = $CACHE_DIR . "cacher_temp.tmp";

my $URLSTR_START = "http://www.cs.helsinki.fi/group/seco/ns/2004/03/18-esinekortti#kortti";
my $COUNT_START = 1;
my $COUNT_END = 239;
my $SEWEHGRIUS_SERVICE = "get_relations_";


##########
# Logic
#

for ( my $i = $COUNT_START; $i < $COUNT_END; $i++ )
{
    my $temp_uri = $URLSTR_START . $i;
    print "URI: $temp_uri\n";

    my $cache_dir_file = $CACHE_DIR . $SEWEHGRIUS_SERVICE
        . my_dummy_url_encode( $temp_uri ) . $CACHE_FILE_SUFFIX;
    print "Cache-file: $cache_dir_file\n";

    my $cmd = $PROLOG_BINARY . " -s " . $PROLOG_SOURCE
        . " -g \"server,ontodella_request_handler:test_relations('$temp_uri'),halt\""
        . " > $TEMP_FILE_NAME";

    system( $cmd );


    # Read the temp file back in and process it

    my $xml = "";
    my $temp_mode = 0;

    open ( TEMPFILE, $TEMP_FILE_NAME ) or
        die "$0: failed to read from file; $!\n";
    while ( my $line = <TEMPFILE> )
    {
        if ( $line =~ m/^<\?xml/ )
        {
            $temp_mode = 1;
        }
        elsif ( $line =~ m|^</rdf:RDF>| )
        {
            $xml .= $line;
            $temp_mode = 0;
        }

        if ( $temp_mode == 1 )
        {
            $xml .= $line;
        }
    }
    close( TEMPFILE );

#    print $xml . "\n\n\n";

    open ( OUTFILE, ">$cache_dir_file" ) or
        die "$0: failed to write to file; $!\n";
    print OUTFILE $xml . "\n";
    close( OUTFILE );
}


sub my_dummy_url_encode
{
    my $unencoded = shift;
    my $encoded = "";

    foreach my $c ( split( //, $unencoded ) )
    {
        if ( $c =~ m/[a-zA-Z0-9 \.\-\*_]/ )
        {
            $encoded .= $c;
        }
        else
        {
            $encoded .= "%" . uc( unpack( "H*", $c ) );
        }
    }

    $encoded =~ tr|[:/%]|[___]|;

#    print "'$encoded'\n";

    return $encoded;
}
