#!/usr/bin/perl -w 

######################################################################
# Creates a test file for checking the rules against the data.

use strict;

if ($#ARGV != 3) {
    die( "Argument missing.\nUSAGE: $0 [prolog] [conf] [prolog test file] [ontodella root]" );
}

my $pl=$ARGV[0];
my $conf=$ARGV[1];
my $prolog_test_file=$ARGV[2];
my $root=$ARGV[3];

# Run prolog for getting the files...

my $cmd = "$pl -t ".
"\"[\'$conf\'], get_config_rdfs_files(F),".
"write(\'<rdffiles>\\n\'),".
"forall(member(File,F),format(\\\"<file>~w</file>~n\\\",File)),".
"write(\'</rdffiles>\\n\'),".
"get_config_rule_files(F2),".
"write(\'<rulefiles>\\n\'),".
"forall(member(File,F2),format(\\\"<file>~w</file>~n\\\",File)),".
"write(\'</rulefiles>\\n\').\"";

my $out = `$cmd 2>/dev/null`;

my @rdffiles = parse_filelist("rdffiles", $out);
my @rulefiles = parse_filelist("rulefiles", $out);

my @uris = ();

foreach my $f (@rulefiles) {
    my $f_content = get_file($f, "prolog");
    my @u = get_uris($f_content, $f);
#    foreach my $uu (@u) {
#        print "UU=$uu 1=".$uu->[0].", 2=".$uu->[1]."\n";
#    }
    push (@uris, @u);
}

my $prolog_out = << "END";
% This file has been created automatically by $0.
% 
% DO NOT MAKE CHANGES TO THIS FILE!

ontodella_test(ErrorCounter) :-
flag(ontodella_test_counter, _, 0),
END
;

for(my $i=0; $i<= $#uris; $i++) {
    my @uu = $uris[$i];
    my $f = $uu[0][0];
    my $u = $uu[0][1];
    if ($i>0) {
	$prolog_out .= ",\n";
    }
    $prolog_out .= << "END";
(
    rdf('$u', _, _);
    rdf(_, '$u', _);
    rdf(_, _, '$u');
    flag(ontodella_test_counter, C$i, C$i+1),
    format('WARNING ~w: \\tData does not contain the uri \\'$u\\' ~n\\t\\tused in rule file \\'$f\\'.~n', [C$i])
END
;

    $prolog_out .= ")";
    $i++;
}

$prolog_out .= << "END"
,
flag(ontodella_test_counter, ErrorCounter, 0),
(
    ErrorCounter>0,
    write('Some potential errors were found.\n\n'),
    write('Please check the rule and data files. Please notice, that
some of the warnings above may be irrelevant. Please consult your
Ontodella expert for more information.\n\n');
    true
).
END
;


open (OUTPUT, ">".$prolog_test_file) or die ("Couldn't write to $prolog_test_file. $!");
print OUTPUT $prolog_out;
close(OUTPUT);

exit;


######################################################################

sub get_file {
    my ($file, $type) = @_;
    open (IN, $file) or die( "couldn't open $file $!" );
    my $cont = "";
    while( my $line=<IN> ) {
        if ($type eq "prolog") {
            $line=~ s/%.*?$//;
	    next if ($line =~ /^\s*?$/);
        }
        $cont .= $line;
    }
    close(IN);
    return $cont;
}

sub get_uris {
    my ($content, $file) = @_;
    my @uris = ();
    while ($content =~ /[\'\"](http:\/\/\S+)[\'\"](\:(\w+)\W)/gis) {
	my $uri = $1;
	if ($3 ne "") {
	    $uri .= $3;
	}
        my @t = [$file, $uri];
	push (@uris, [$file, $uri]);
    }
    return @uris;
}


sub parse_filelist {
    my ($filelist, $out) = @_;
    $out =~ /<$filelist>(.*?)<\/$filelist>/sig;
    my $files_t = $1;
    my @files = ();
    while ($files_t =~ /<file>(.*?)<\/file>/gi) {
        push(@files, $root.$1);
    }

    return @files;
}
