/*  $Id: ontodella_request_handler.pl,v 1.48 2004/03/18 13:48:51 kimvilja Exp $

    Part of OnToDella

    Author:        Arttu Valo
    E-mail:        arttu.valo@helsinki.fi
    WWW:           http://www.cs.helsinki.fi/group/seco

Copyright (c) 2003-2004 Helsinki Institute for Information Technology

Permission is hereby granted, free of charge, to any person obtaining 
a copy of this software and associated documentation files (the "Software"), 
to deal in the Software without restriction, including without limitation 
the rights to use, copy, modify, merge, publish, distribute, sublicense, 
and/or sell copies of the Software, and to permit persons to whom the 
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included 
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS 
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN 
THE SOFTWARE.
*/

:- module( ontodella_request_handler,
	  [ reply/1
	  ]).
:- use_module(library('http/http_client')).

:- style_check(-atom).			% allow long atoms

% load SeWeHGrius rules, first the core, then the configured rules
:- ontodella_file( 'lib/prolog/sewehgrius_rdfs_core_rules.pl', SwehgRules ),
   load_files( SwehgRules, [imports(all)] ).

%:- get_config_load_rules_files.
%:- 
%   %get_config_recommendation_logic_file( RecommLogicFile ),
%
%   %%%% HARDCODING...
%
%   ontodella_file( 'src/prolog/simple_rdfs_distance_recommender.pl', RecommLogicFile ),
%    load_files( RecommLogicFile, [imports(all)] ).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% must be defined before calling...

load_prolog_file( File ) :-
%    format( 'Load prolog file: ~w~n', [File] ),
    ontodella_file( File, LoadFile ),
    load_files( LoadFile, [imports(all)] ).


:- 
    get_config_rule_files( RuleFiles ),
    checklist( load_prolog_file, RuleFiles ).



% reply predicate which wraps everything inside a catch. Does this work???
%
reply( Request ) :-
    catch(reply2(Request), Exception, 
    (
       sformat(ErrStr, 'Error: ~w', [Exception]),
       print_out_http_fail( 500, ErrStr ),
       writelog( 'Exception: ~w', [ErrStr] )
    )),
    http_workers(Port,Workers),
    writelog('workers: port=~w, workers=~w', [Port,Workers]),
    !.


% write the request to the log and continue...
reply2( Request ) :-
    member( method(Method), Request ),
    writelog( 'got request: ~w, METHOD=~w', [Request, Method] ),
    fail.


% reply predicate for reloading the RDF(S) files
reply2( Request ) :-
    member( path( Path ), Request ), % get the path
    Path = '/reloadRDFS',
    format( 'Content-type: text/plain~n~n', [] ), % TODO: this should use the ontodella header!
    format( 'Path was ~w~n\n', [Path] ),
    format( 'So reloading the RDF(S) files was requested.\n' ),
    format( 'UNLOADING ...\n' ),
    catch(
	(
            call( unload_rdfs_files ),
	    format( 'Unloaded.\nRELOADING ...\n' )
        ),
	Exception,
	(
	    format( "FAILED: ~w\n", [Exception] )
	)
    ),
    catch(
        (
            call( load_rdfs_files ),
	    format( 'Reloaded.\nHave a nice day.\n' )
        ),
	Exception2,
	(
	    format( "FAILED: ~w\n", [Exception2] )
	)
    ),
    !.

% Predicate for echoing request contents
reply2( Request ) :-
    member( path( Path ), Request ),
    Path = '/debugecho',
    writelog( 'called: debug echo' ),
    format( 'HTTP/1.1 200 OK\nContent-Type: text/plain\n\n' ), % TODO: this should use the ontodella header
    format( 'DEBUG ECHO:\n' ),
    get_http_params( Request, Params ),
    format( '~w\n', [Params] ).
%    writelog( '2: ~w~n', [Request] ),
%    (
%        member( method(post), Request ),
%        http_read_data( Request, Params, [] ),
%writelog( '3' ),
%        format( 'Params:\n~w\n', [Params] );
%writelog( '4' ),
%        true
%    ).

% reply predicate for dynamic predicate service calling
reply2( Request ) :-
    !,  % don't try other service methods either if this fails
    member( path( Path ), Request ),  % get the path
    (
        get_http_params( Request, Params );
        Params = Request
    ),
    catch(
        (
            reply_core( Path, Params ),
%            print_out_http_ok,
            StatusCode = 200
        ),
	Exception,
        (
            (
                Exception = http_status( StatusCode, ErrorStr );
                StatusCode = 500,
                term_to_atom( Exception, ExceptionAtom ),
                ErrorStr = ExceptionAtom
            ),
	    print_out_http_fail( StatusCode, ErrorStr )
        )
    ),
%    (
%        get_config_use_logging,
%        string_concat( StatusCode, ' ', TempStr1 ),
%        string_concat( TempStr1, Path, ResultStr ),
	writelog( '~w ~w ~w', [StatusCode, Path, Params] ).
%
%        true % why is this needed?
%
%   ).  % if no logging, just merrily skip it and be happy

% last reply predicate: if none of the above matched, file a complaint at the office
reply2( _ ) :-
    print_out_http_fail( 400, 'Really did not know what to do with your request.' ).


% Gets the "hash-table" of parameters associated with the request.
%get_http_params( +Request, -Params )
get_http_params( Request, Params ) :-
    nonvar( Request ),
    member( method( post ), Request),
    http_read_data( Request, Params, [] ).

get_http_params( Request, Params ) :-
    nonvar( Request ),
    member( method( get ), Request),
    Params=Request.   % AV: WRONG? FIXXX? We only use POST anyway...

get_http_params( A, B ) :-
writelog( 'A=~w, B=~w~n~n', [A,B] ),
    throw( http_status( 400, 'Problem with your request type or parameters. Only HTTP POSTs certainly supported. (get_http_params)') ).


%reply_core( +Path, +Data )
reply_core( Path, Data ) :-
    !,  % Only try this once, cut all other tries!
    % See, which service is requested
    get_predicate_name_portion_of_url( Path, PredicateName ),
    secure_predicate_name( PredicateName ),
    string_to_atom( PredicateName, PredicateAtom ),
    initialize_response_buffer,
    !,
    catch(
        (
	    call( PredicateAtom, Data )
	),
	Exception,
	(
            string_to_atom( 'Internal server error: ', ISEAtom ),
            term_to_atom( Exception, ExceptionAtom ),
	    string_concat( ISEAtom, ExceptionAtom, Errorstr ),
	    throw( http_status( 500, Errorstr ) )
	)
    ).


%%%%%%%%%%
% Printing out HTTP responses and status codes
%

%print_out_http_ok
print_out_http_ok :-
    print_out_http_status_line( 200, 'OK' ),
    format( "Content-Type: text/plain\n" ),
%    format( "Content-Type: text/plain; charset=\"iso-8859-1\"\n" ),
    get_thread_output_buffer_id( BufferId ),
    recorded( BufferId, Value, Reference ),  % "hash-table-access"
    format( "~w\n", [ Value ] ),
    erase( Reference ).  % clear the memory before serving next request


%print_out_http_fail( +Code, +Message )
print_out_http_fail( Code, Message ) :-
    print_out_http_status_line( Code, Message ),
    write( '\n' ),
    format( "~w: ~w", [Code, Message] ).    % printing the error also to the user (debug purposes)


%print_out_http_status_line( +Code, +Message )
print_out_http_status_line( Code, Message ) :-
    string_to_atom( Code, CodeAtom ),
    string_to_atom( Message, MessageAtom ),
    format( "HTTP/1.1 ~w ~w~n", [CodeAtom, MessageAtom] ),
    format( "Server: OntoDella Prolog HTTP Server/1.0~n" ).


%%%%%%%%%%
% Printing into the response 'buffer'
%


%initialize_response_buffer
initialize_response_buffer :-
    get_thread_output_buffer_id( BufferId ),
    recorda( BufferId, '' ).

%get_thread_output_buffer_id( BufferId )
get_thread_output_buffer_id( BufferId ) :-
    thread_self( ThreadId ),  % get the thread id
    string_concat( ThreadId, 'respbuffer', OutputBufferName ),
    string_to_atom( OutputBufferName, BufferId ).


%http_print( String )
http_print( Term ) :-
    !, % never let this back up!
    (
        is_list( Term ),
        throw( http_status( 500, 'Internal server error: http_print cannot print lists!' ) );

        atomic( Term ),
        PrintString = Term;

        term_to_atom( Term, PrintString );

        throw( http_status( 500, 'Internal server error: http_print cannot handle this type!' ) )
    ),
% AV: internal secondary buffering removed, prolog's default
% httpd-wrapper already uses a memory file (although this renders our
% own exception handling useless and even wrong ...)
%
%    get_thread_output_buffer_id( BufferId ),
%    (
        % Check what's in the buffer (if anything or even whether it exists)
%        recorded( BufferId, Value, Reference ),
        % if the buffer has something, append to it
        % check that the buffer existed, otherwise backup to the other branch
%        nonvar( Reference ),
%        erase( Reference ),  % SLOW'N'UGLY, but tell me how to solve this better!
%        string_concat( Value, PrintString, ConcattedString );

        % the buffer was not initialized, let's do it
%        initialize_response_buffer,
%        ConcattedString = Value
%    ),
%    recorda( BufferId, ConcattedString ).

    encode_ampersands( PrintString, EncodedPrintString ), 
    format( '~w', [EncodedPrintString] ).


%%%%%%%%%
% DYNAMICALLY CALLABLE SERVICE METHODS (OFFICIAL ONES)
%


% testing predicates for dynamic calling
foobar( _ ) :-
    http_print( 'You have called the predicate named FOOBAR!<br/>\n' ).
thou_shalt_not_pass( _ ) :-
    http_print( 'Gandalf the Rainbowy: Thou shalt not pass without proper alignment!<br/>\n' ).


% Gets the triplets associated with a given uri (object, predicate or subject)
%get_annotations( +Params )
get_annotations( Params ) :-
    get_http_parameter( Params, uri, URIString ),
    normal_string_to_term( URIString, URI ),
    print_out_http_ok,

    http_print_rdf_doctype,
    http_begin_rdf_document,

    foreach( sewehgrius_annotation( URI, Namespace, Name, Value ),
        http_print_rdf_annotation( URI, Namespace, Name, Value ) ),

    http_end_rdf_document,
    !.


% Gets the categories, no parameters are currently needed
%get_categories( +Params )
get_categories( _ ) :-
    print_out_http_ok,  % no more buffering for the gargantuan data this creates
    http_print_rdf_doctype,
    http_begin_rdf_document,

    % zero the root category id counter
    flag( 'rootcatid', _, 0 ),

	writef('<rdf:Description rdf:about="&ogt;RootCategory">\n<ogt:subCategories rdf:parseType="Collection">\n'),
    foreach(
        sewehgrius_category( RootURIAtom, LabelPred, SubStructuringPredicateAtom, InstancePredicateAtom ),
        print_category( RootURIAtom, LabelPred, SubStructuringPredicateAtom, InstancePredicateAtom, root )
    ),  % foreach
    writef('</ogt:subCategories>\n</rdf:Description>\n'),

    % append the actual bookmarks regardles of whether they appear
    % in the categories above

    foreach(
        esinekortti( SubjectURI ),  % FIXXX! HARDCODING!
        http_print_bookmark( SubjectURI )
    ),

    http_end_rdf_document,
    !.


% Gets the relations (~links) with the parameters given
%get_relations( +Params )
get_relations( Params ) :-

%% Print to log

    writelog( "calling: get_relations with ~w", [Params]),

    get_rdf_triple_parameters( Params, Subject, Predicate, Object ),

    print_out_http_ok,
    http_print_rdf_doctype,
    http_begin_rdf_document,

    % for multiple URIs, the loop should start here

    http_print( '<rdf:Description rdf:about=' ),
    http_print_quoted_string( Subject ),
    http_print( '>\n' ),
    http_print( '<swhg:relations rdf:parseType="Collection">\n' ),

    % for each defined interesting and amazing relation rule, get its 
    % pseudo-URI, text equivalent labels list and predicate which
    % implements the sewehgrius-api and thus returns the information
    % of items (uris) which match for the relation (i.e., succeed)

    setof(
        TempResultSet,
        get_relations_resultset( [], TempResultSet, Subject, Predicate, Object ),
        ResultSetSet
    ),

    TempResultSetSet = [ResultSetSet|_],

    collapse_resultsetset( TempResultSetSet, _, FullResultList ),
%    writelog( 'get relations: TempResultSetSet=~w~n----------~nFullResultList=~w~n~n', [TempResultSetSet, FullResultList] ),
    garbage_collect,  % try to optimize and limit memory usage

    % Process the acquired result set
    % It contains an uncomfortable almost-tree-like structure
    % full of quadruples [A, B, C, D] none of which should be lost.

    merge_entire_listtree( FullResultList, FullResultList2 ),

%    writelog( 'get relations: FullResultList2=~w~n~n', [FullResultList2] ),

    get_distinct_uri_pairs( FullResultList2, DistinctPairList ),

%    writelog( 'get relations DistinctPairList=~w~n~n', [DistinctPairList] ),

    calculate_resultset_weights( DistinctPairList, ScoreList ),

%    writelog( 'get relations ScoreList=~w~n~n', [ScoreList] ),

    garbage_collect,  % try to optimize and limit memory usage

    % FIXXX: sorting could be done here (but is left to the XSLT)

    % Gotta fool around with these lists :(
    FullResultList = [AllResults|_],
    http_print_relations_resultset( AllResults, ScoreList ),

    http_print( '</swhg:relations>\n' ),
    http_print( '</rdf:Description>\n' ),

    % the multi-URI loop should end here

    http_end_rdf_document,
    !.


%%%%%%%%%%
% UTILITY PREDICATES
%


%%%%%
% XML/RDF/HTML Printing (into the 'buffer')
%


%http_print_rdf_doctype
http_print_rdf_doctype :-
    write( '<?xml version="1.0" encoding="ISO-8859-1"?>
<!DOCTYPE rdf:RDF [
 <!ENTITY bm "http://www.w3.org/2002/01/bookmark#">
 <!ENTITY ogt "http://www.cs.helsinki.fi/group/seco/ns/2004/03/ontogator#">
 <!ENTITY rdf "http://www.w3.org/1999/02/22-rdf-syntax-ns#">
 <!ENTITY rdfs "http://www.w3.org/2000/01/rdf-schema#">
 <!ENTITY swhg "http://www.cs.helsinki.fi/group/seco/ns/2004/03/18-sewehgrius#">
 <!ENTITY fms "http://www.cs.helsinki.fi/group/seco/ns/2004/03/18-esinekortti#">
 <!ENTITY masa "http://www.cs.helsinki.fi/group/seco/ns/2004/03/18-masa#">
 <!ENTITY mao "http://www.cs.helsinki.fi/group/seco/ns/2004/03/18-mao#">
 <!ENTITY maoryhma "http://www.cs.helsinki.fi/group/seco/ns/2004/03/18-maoryhma#">
 <!ENTITY terms "http://www.cs.helsinki.fi/group/seco/ns/2004/03/18-terms#">
 <!ENTITY ajat "http://www.cs.helsinki.fi/group/seco/ns/2004/03/18-ajat#">
 <!ENTITY kokoelmat "http://www.cs.helsinki.fi/group/seco/ns/2004/03/18-kokoelmat#">
 <!ENTITY owl "http://www.w3.org/2002/07/owl#">
 <!ENTITY od "http://www.cs.helsinki.fi/group/seco/ns/2004/03/18-ontodella#">
]>
' ).
% Old RDFS schema UR[L|I|N] ending: TR/1999/PR-rdf-schema-19990303#">


%http_begin_rdf_document
http_begin_rdf_document :-
    sewehgrius_metadata('dataid', DataID),

    format( '<rdf:RDF xmlns:bm="&bm;" xmlns:ogt="&ogt;" xmlns:rdf="&rdf;"
 xmlns:rdfs="&rdfs;" xmlns:swhg="&swhg;" xmlns:fms="&fms;" 
 xmlns:mao="&mao;" xmlns:masa="&masa;" xmlns:maoryhma="&maoryhma;" 
 xmlns:terms="&terms;" xmlns:ajat="&ajat;" xmlns:kokoelmat="&kokoelmat;"
 xmlns:owl="&owl;" xmlns:od="&od;">
<!-- metadata about ontodella in ontogator format -->
<ogt:CategoryBase rdf:about="">
  <owl:versionInfo>~w</owl:versionInfo>
  <od:ontodellaCvsId>$Id: ontodella_request_handler.pl,v 1.48 2004/03/18 13:48:51 kimvilja Exp $</od:ontodellaCvsId>
', [DataID] ),
    gethostname( HostName ),
    get_time(T), 
    convert_time(T, DateStr),
    forall( (sewehgrius_metadata(ID, Data), format('  <od:sewehgrius_~w>~w</od:sewehgrius_~w>~n', [ID,Data,ID])), true ),
    format( "  <od:replyCreationTime>~w</od:replyCreationTime>~n", [DateStr] ),
    format( "  <od:ontodellaHost>~w</od:ontodellaHost>~n", [HostName] ),
    write( '</ogt:CategoryBase>\n' ).

%http_end_rdf_document
http_end_rdf_document :-
    http_print( '</rdf:RDF>\n' ).


% Safely prints out a string surrounded by suitable quotation marks
%http_print_quoted_string( +Printable )
http_print_quoted_string( Printable ) :-
    !,  % newer let this predicate back up from here
    (
        string( Printable ),
        string_to_atom( Printable, Atom );

        atomic( Printable ),
        Atom = Printable;

        Printable = Namespace:ID,
        term_to_atom( Namespace, NsAtom ),
        string_to_atom( NsString, NsAtom ),
        term_to_atom( ID, IdAtom ),
        string_to_atom( IdString, IdAtom ),
        string_concat( NsString, IdString, String3 ),
        string_to_atom( String3, Atom );

        term_to_atom( Printable, Atom );

%        writef( 'ERRORstring: %w :ERRORstring\n', [Printable] ),
%        term_to_atom( Printable, Atom ),
%        writef( 'As atom: "%w"\n', [Atom] ),

        throw( http_status( 500, 'Internal server error: http_print_quoted_string cannot handle this data type correctly!') )

    ),
    !,  % don't let things back up here either (problems with different types!)
    (   % if there is a " in the string, quote with 's, otherwise with "s
        sub_string( Atom, _, _, _, '\"' ),  % efficiency!?
        Quote = '\'';

        Quote = '\"'
    ),
    http_print( Quote ),
    http_print( Atom ),
    http_print( Quote ).


% Prints out rdfs:label tags with respective xml:lang attributes and texts
%http_print_xmllang_list( LabelList )
http_print_xmllang_list( [] ) :- true.

http_print_xmllang_list( [Language:Text|LabelList] ) :-
    http_print( '<rdfs:label xml:lang=' ),
    http_print_quoted_string( Language ),
    http_print( '>' ),
    http_print( Text ),
    http_print( '</rdfs:label>\n' ),
    http_print_xmllang_list( LabelList ).  % recursively print out the remainder of the list

http_print_xmllang_list( [Text|LabelList] ) :-
    http_print( '<rdfs:label>' ),
    http_print( Text ),
    http_print( '</rdfs:label>\n' ),
    http_print_xmllang_list( LabelList ).  % recursively print out the remainder of the list

%http_print_start_collection( +Tag )
http_print_start_collection( Tag ) :-
    http_print( '<' ),
    http_print( Tag ),
    http_print( ' parseType="Collection">\n' ).


%http_print_end_collection( +Tag )
http_print_end_collection( Tag ) :-
    http_print( '</' ),
    http_print( Tag ),
    http_print( '>\n' ).


%http_print_rdf_annotation( +URI, +Namespace, +Name, +Value )
http_print_rdf_annotation( URI, Namespace, Name, Value ) :-
    http_print( '<rdf:Description rdf:about=' ),
    http_print_quoted_string( URI ),
    http_print( '>\n<rdf:predicate rdf:resource="' ),
    http_print( Namespace ),
    http_print( Name ),
    http_print( '"/>\n<rdfs:label>' ),
    (
        Value = literal( DeLiteralizedValue ),   % if it is a literal ...
        http_print( DeLiteralizedValue );

        http_print( Value )
    ),
    http_print( '</rdfs:label>\n<swhg:name>' ),
    http_print( Name ),
    http_print( '</swhg:name>\n</rdf:Description>\n' ).


%http_print_start_rule( +RelationURIAtom, +RelationLabelList )
http_print_start_rule( RelationURIAtom, RelationLabelList ) :-
    http_print( '<swhg:RelationGroup>\n' ),
    http_print( '<swhg:relationDescription rdf:resource=' ),
    http_print_quoted_string( RelationURIAtom ),
    http_print( ' />\n' ),
    http_print_xmllang_list( RelationLabelList ).


%http_print_end_rule
http_print_end_rule :-
    http_print( '</swhg:RelationGroup>\n' ).


%http_print_start_relinstance( +RelDesc )
http_print_start_relinstance( RelDesc ) :-
    http_print( '<swhg:relations rdf:parseType="Collection">\n' ),
    http_print( '<swhg:RelationGroup>\n' ),
    http_print_relation_description( RelDesc, 'label' ),
    http_print( '<swhg:relationDescription rdf:parseType="Collection">\n' ),
    http_print_relation_description( RelDesc, not:'label' ),
    http_print( '</swhg:relationDescription>\n' ),
    http_print( '<swhg:relations rdf:parseType="Collection">\n' ).

%http_print_end_relinstance/0
http_print_end_relinstance :-
    http_print(' </swhg:relations>\n' ),
    http_print( '</swhg:RelationGroup>\n' ),
    http_print( '</swhg:relations>\n' ).

%http_print_relation_description( +RelDesc )

http_print_relation_description( List ):-
    http_print_relation_description( List, '' ).

http_print_relation_description( [], _ ).

http_print_relation_description( Desc, Type ):-
    (
        is_list(Desc),
        forall(member(Member, Desc), 
            ((
                (
                    Type\='',
                    functor(Member, Type, _);
                    Type=not:NType,
                    functor(Member, FType, _),
                    NType\=FType;
                    Type=''
                ),
                (
                    nonvar( Member ),
                    http_print_relation_description_member( Member );
                    true
                ));
                true
            )
        );
        not(is_list(Desc)),
        http_print_relation_description_member( Desc )
    );
    true
    .

http_print_relation_description_member( commonResources(CommonResources) ) :-
    (
        is_list(CommonResources),
        forall(member(Res, CommonResources),
        (
            http_print_relation_description_commonResource( Res )
        ))
    );
    (
	http_print_relation_description_commonResource( CommonResources )
    ).

http_print_relation_description_member( label(Labels) ) :-
    (
        is_list( Labels ),
        LabelsL = Labels;
        LabelsL = [Labels]
    ),
    http_print_xmllang_list( LabelsL ).


http_print_relation_description_commonResource( literal(Literal) ) :-
    http_print( '<swhg:commonResource>\n<rdfs:label>' ),
    http_print( Literal ),
    http_print( '</rdfs:label>\n</swhg:commonResource>\n' ).

http_print_relation_description_commonResource( Resource ) :-
    http_print( '<swhg:commonResource rdf:about=' ),
    http_print_quoted_string( Resource ),
    http_print( '>\n'),

    % try to get labels
    (
        sewehgrius_labels( Resource, Labels ),
        http_print_xmllang_list( Labels );

        true
    ),

    http_print( '</swhg:commonResource>\n' ).



http_print_relation_description( RelDesc ) :-
    http_print( 'Couldn\'t print this nicely...' ),
    sformat( Str, '~w', RelDesc ),
    http_print( Str ).

%http_print_relations_resultset( +ResultSet, +ScoreList )
http_print_relations_resultset( [], _ ).

http_print_relations_resultset( [[]|Tail], ScoreList ) :-
    % empty set in the middle: no worries, just skip to next
    http_print_relations_resultset( Tail, ScoreList ).

http_print_relations_resultset( [RuleResults|Tail], ScoreList ) :-
    % get the Rule URI (from the first instance 
    RuleResults = [[RuleURI, _, _, _ ]|_],
    % get its labels
    sewehgrius_relation_rule( RuleURI, RuleLabels, _ ),
    % print opening tag
    http_print_start_rule( RuleURI, RuleLabels ),
    % zero the counter
    get_thread_rule_relation_counter_id( CounterName ),
    flag( CounterName, _, 0 ),
    % print substructure
    get_distinct_relinstance_uris( RuleResults, DistinctItems ),
    http_print_distinct_relinstances( DistinctItems, RuleResults, ScoreList ),
    % close the tag
    http_print_end_rule,
    garbage_collect,  % try to optimize and limit memory usage
    http_print_relations_resultset( Tail, ScoreList ).


%http_print_distinct_relinstances( DistinctItemList, RelationList, ScoreList )
http_print_distinct_relinstances( [], _, _ ).

http_print_distinct_relinstances( [RelTarget|Tail], RelationList, ScoreList ) :-
    http_print_start_relinstance( RelTarget ),
    http_print_relinstance( RelTarget, RelationList, ScoreList ),
    http_print_end_relinstance,
    http_print_distinct_relinstances( Tail, RelationList, ScoreList ).


%http_print_relinstance( +RelInstanceURI, +RelationList, +ScoreList )
http_print_relinstance( _, [], _ ).

http_print_relinstance( RelInstanceURI, [Item|Tail], ScoreList ) :-
    % if the item contains same "relation instance",
    % print it; otherwise happily proceed to next one
    (
        Item = [_, RelInstanceURI, _, _],
        http_print_relation_item( Item, ScoreList );

        true
    ),
    http_print_relinstance( RelInstanceURI, Tail, ScoreList ).


% Actually prints a relation item, its labels (if any are found) and its
% weight/score (or -1 if none is found)
%http_print_relation_item( ItemStructure, ScoreList )
http_print_relation_item( [], _ ).

http_print_relation_item( [_, _, SourceURI, TargetURI], ScoreList ) :-
    % finally we are actually printing something, so we increment the counter
    get_thread_rule_relation_counter_id( CounterName ),
    flag( CounterName, Count, Count+1 ),

    % check that we are within the limits given to us
    get_config_item_count_limit( Limit ),

    % only print if within limits, but succeed anyway
    (
        Count < Limit,

        % get the weight -- or if it does not exist (e.g., annotation!?)
        % use constant
        (
            member( [SourceURI, TargetURI, Weight], ScoreList );

%	    writelog( 'http_print_relation: returning -1, since [~w, ~w, ~w] is not in ScoreList( ~w ).~n', [SourceURI, TargetURI, Weight, ScoreList] ),
            Weight = -1
        ),
        http_print( '<swhg:relation swhg:target=' ),
        http_print_quoted_string( TargetURI ),
        http_print( ' swhg:weight=' ),
        http_print_quoted_string( Weight ),
        http_print( '>\n' ),

        % try to get labels
        (
            sewehgrius_labels( TargetURI, Labels ),
            http_print_xmllang_list( Labels );

            true
        ),
        http_print( '</swhg:relation>\n');

        % limit exceeded, no printing, just succeeding
        true
    ),
    trim_stacks.


%http_print_bookmark( +URI )
http_print_bookmark( URI ) :-
    http_print( '<bm:Bookmark rdf:about=' ),
    http_print_quoted_string( URI ),
    http_print( '>\n' ),

    http_print_projectionof( URI ),
    sewehgrius_labels( URI, Labels ),
    http_print_xmllang_list( Labels ),

    get_annotation_list( URI, Annotations ),
    http_print_annotations( Annotations ),

    http_print( '</bm:Bookmark>\n' ).


%http_print_annotation( AnnotationSet )
http_print_annotations( [] ).

http_print_annotations( [[Namespace, Name, literal( Value )]|Tail]  ) :-
    % value was a literal
    http_print_annotation_opener_tag( Namespace, Name ),
    encode_gt_lt( Value, EValue ),
    http_print( EValue ),
    http_print_annotation_closer_tag( Namespace, Name ),
    http_print_annotations( Tail ).

http_print_annotations( [[Namespace, Name, Value]|Tail] ) :-
    % value was not a literal, i.e., it was a resource (URI)
    http_print_annotation_opener_tag( Namespace, Name ),
    http_print( '\n<rdf:Description rdf:about=' ),
    http_print_quoted_string( Value ),
    http_print( '>\n' ),

    % get some label for the value URI (resource)
    sewehgrius_labels( Value, Labels ),
    http_print_xmllang_list( Labels ),

    http_print( '</rdf:Description>\n' ),
    http_print_annotation_closer_tag( Namespace, Name ),
    http_print_annotations( Tail ).


%http_print_annotation_opener_tag( +Namespace, +Name )
http_print_annotation_opener_tag( Namespace, Name ) :-
    http_print( '<' ),
    (
        nonvar( Namespace ),  % don't match to an unbound variables
        Namespace = rdf,
        http_print( rdf ),
        http_print( ':' );

        % otherwise we catch the annotations into our own namespace
        http_print( 'fms:' )
    ),
    http_print( Name ),
    http_print( '>').

%http_print_annotation_closer_tag( +Namespace, +Name )
http_print_annotation_closer_tag( Namespace, Name ) :-
    http_print( '</' ),
    (
        nonvar( Namespace ),  % don't match to an unbound variable
        Namespace = rdf,
        http_print( rdf ),
        http_print( ':' );

        % otherwise we catch the annotations into our own namespace
        http_print( 'fms:' )
    ),
    http_print( Name ),
    http_print( '>\n' ).


%%%%
% Miscellaneous
%


%get_annotation_list( +URI, -Annotations )
get_annotation_list( URI, Annotations ) :-
    setof(
        [Namespace, Name, Value],
        sewehgrius_annotation( URI, Namespace, Name, Value ),
        Annotations
    );  % or!
    Annotations = [].


% Gets the thread's counter id for the rule relation count limiting
%get_thread_rule_relation_counter_id( -CounterName )
get_thread_rule_relation_counter_id( CounterName ) :-
    thread_self( ThreadId ),  % get the thread id
    string_concat( ThreadId, 'rulerelcount', NameString ),
    string_to_atom( NameString, CounterName ).


%collapse_resultsetset( +ResultSetSet, -PresentList, -FullResultList )
collapse_resultsetset( [], [], _ ).

collapse_resultsetset( [ASet|Tail], PresentList, FullResultList ) :-
    collapse_resultsetset( Tail, SubPresentList, FullResultList ),
    sort( ASet, SortedSet ),

    % if it is an empty set, forget it; otherwise merge

    (
        SortedSet = [],
        FullResultList = SubPresentList,
        writeln( 'Was empty set' );

        merge_set( SortedSet, PresentList, FullResultList )
    ).



%get_relations_resultset( +OldResultSet, -NewResultSet, +Subject, +Predicate, +Object )
get_relations_resultset( OldResultSet, NewResultSet, Subject, Predicate, Object ) :-

    % collect resultsets for all rules into one list,
    % then collapse the top level of that list (append, not merge)

    findall(
        RuleResultSet,

        (
            sewehgrius_relation_rule(
                RuleURI, _, RulePredicate
            ),

            % get_rule_resultset also works down any existing rule
            % hierarchy

            get_rule_resultset(
                OldResultSet, RuleResultSet, RuleURI, RulePredicate,
                Subject, Predicate, Object
            )
        ), % 2nd branch of findall

        NewResultSet
    ).  % findall


%get_rule_resultset( -OldResultSet, -NewResultSet, +RuleURI, +RulePredicate, ?Subject, ?Predicate, ?Object )
get_rule_resultset( OldResultSet, NewResultSet, RuleURI, RulePredicate,
    Subject, _, _ ) :-    % Predicate, Object ) :-  <-- these params currently not in use

    % FIXXX: the sub-rules should be handled here as well!?

    findall(
        [RuleURI, RelationDescription, Subject, Object],
        call( RulePredicate, Subject, RelationDescription, Object ),
        MyResultSet
    ),
    sort( MyResultSet, MySortedResultSet ),
    merge( OldResultSet, MySortedResultSet, NewResultSet).


%get_distinct_relation_uris( +ResultList, -RelationList )
get_distinct_relation_uris( [], [] ).

get_distinct_relation_uris( [[_, RelationURI, _]|ResultList], RelationList ) :-
    get_distinct_relation_uris( ResultList, SubRelationList ),
    % if it is not already there, add the RelationURI to the RelationList
    (
        member( RelationURI, SubRelationList ),
        RelationList = SubRelationList;

        RelationList = [RelationURI|SubRelationList]
    ).


%get_distinct_relinstance_uris( +ItemList, -DistinctItems )
get_distinct_relinstance_uris( [], [] ).

get_distinct_relinstance_uris( [[_, RelInstURI, _, _]|Tail], DistinctItems ) :-
    get_distinct_relinstance_uris( Tail, SubDistinctItems ),
    (
        member( RelInstURI, SubDistinctItems ),
        DistinctItems = SubDistinctItems;

        DistinctItems = [RelInstURI|SubDistinctItems]
    ).


%get_distinct_uri_pairs( +FullResultList, -DistinctPairList )
get_distinct_uri_pairs( [], [] ).

get_distinct_uri_pairs( [[_, _, SubjectURI, ObjectURI]|Tail], DistinctPairList ) :-
    get_distinct_uri_pairs( Tail, SubResultList ),
    % if (Subj, Obj) is not already in the list, add it
    (
        member( [SubjectURI, ObjectURI], SubResultList ),
        DistinctPairList = SubResultList;

        DistinctPairList = [[SubjectURI, ObjectURI]|SubResultList]
    ).

% for cases where list contains empty result lists
get_distinct_uri_pairs( [[]|Tail], DistinctPairList ) :-
    get_distinct_uri_pairs( Tail, DistinctPairList ).


%%%%%%%%%%
% Score/Weight calculation
%


%calculate_resultset_weights( +FullResultSet, -ScoreList )
calculate_resultset_weights( [], [] ).

calculate_resultset_weights( [[Subject, Object]|Tail], ScoreList ) :-
    calculate_resultset_weights( Tail, TailScoreList ),

    % If recommending is configured to be used, spend the time calculating

    (
        get_config_use_recommending,
        calculate_uripair_score( Subject, Object, Weight );

        Weight = 1
    ),
    ScoreList = [[Subject, Object, Weight]|TailScoreList].


%calculate_uripair_score( +SubjectURI, +ObjectURI, -Score )
calculate_uripair_score( SubjectURI, SubjectURI, 1 ) :- !. % prune the simple case



% get the uripair from the uripair cache.
calculate_uripair_score( SubjectURI, ObjectURI, Score ) :-
	get_from_uripair_cache( SubjectURI, ObjectURI, Score ),
	!.


% calculate the uripair score
calculate_uripair_score( SubjectURI, ObjectURI, Score ) :-

    % set up the counters

    thread_self( ThreadId ),  % get the thread id
    string_concat( ThreadId, 'subjectpropertycount', SubjectCountFlagName ),
    string_to_atom( SubjectCountFlagName, SubjectCountFlag ),
    string_concat( ThreadId, 'objectpropertycount', ObjectCountFlagName ),
    string_to_atom( ObjectCountFlagName, ObjectCountFlag ),
    flag( SubjectCountFlag, _, 0 ),
    flag( ObjectCountFlag, _, 0 ),

    % for each property

    findall(
        [ClassOfType1, ClassOfType2, Path, Distance, Score ],  % only Score used, others are filler
        (
            (
                rdf( SubjectURI, Predicate1, Type1 ),
                Type1 \= literal( _ ),
                flag( SubjectCountFlag, Old1, Old1+1 );

                true
            ),
            (
                rdf( ObjectURI, Predicate2, Type2 ),
                Type2 \= literal( _ ),
                flag( ObjectCountFlag, Old2, Old2+1 );

                true
            ),
            nonvar( Predicate1 ),
            Predicate1 = Predicate2,
            % note: if Type1 or Type2 is a class, we should just keep them as their own type
	    % WARNING: get_type/2 is declared in simple_rdfs_distance_recommender.pl
            get_type( Type1, ClassOfType1 ),
            get_type( Type2, ClassOfType2 ),
            distance( ClassOfType1, ClassOfType2, Path, Distance, Score );

            Score = 0
        ),
        ScoreList
    ),

    % find the numbers of the properties assigned to (lengths/size)

    % get the flags
    flag( SubjectCountFlag, SubjectPropertyCount, SubjectPropertyCount ),
    flag( ObjectCountFlag, ObjectPropertyCount, ObjectPropertyCount ),
    LargerResultSize is max( SubjectPropertyCount, ObjectPropertyCount ),

    predsort( distance_result_compare, ScoreList, SortedScoreList ),
    avg_res( SortedScoreList, LargerResultSize, AvgWeight ),
%writef( '%w weight with %w is %w\n', [SubjectURI, ObjectURI, AvgWeight] ),
    Score = AvgWeight,


    add_to_uripair_cache( SubjectURI, ObjectURI, Score ).



% Creates a local category file hard copy for deployment.
%get_categories_to_temp_file/0
get_categories_to_temp_file :-
    % create the darn gargantuan categories file
    writeln( 'Producing category file, please wait, this might take minutes!' ),
    current_output( OldStream ),
    open( 'get_categories_.tmp', write, -CategoryStream ),
    set_output( CategoryStream ),
    get_categories( [] ),
    close( CategoryStream ),
    set_output( OldStream ),
    writeln( 'Category file produced.' ).


% Checks whether the predicate to be called dynamically is one of the reportedly
% allowed ones. At least predicat "shell" should definitely be off limits.
%secure_predicate_name( +PredicateNameString )
secure_predicate_name( PredicateNameString ) :-
    (
        string( PredicateNameString )->string_to_atom( PredicateNameString, PredicateAtom );  % turn the name string into an atom
        atom( PredicateNameString )->PredicateAtom = PredicateNameString
    ),
    member( PredicateAtom, [ get_annotations,
                             get_categories,
                             get_categories2,
                             get_relations,
                             get_nearest,
                             get_nearest_muinainen,
			     foobar,
			     thou_shalt_not_pass] ).

secure_predicate_name( _ ) :-
    throw( http_status(400, 'Blocked by Security: bad predicate requested') ).


%%%%%%%%
% Generic RDF Handling
%


%%%%
% Inbound (HTTP) parameter fixing
%


%get_http_parameter( +Params, +Name, -Value ),
get_http_parameter( Params, Name, Value ) :-
    ( member( Name=Value, Params ), !; true, ! ). % true: leave the variable as untied


% Gets the rdf triple parameters (subject, object, predicate) and if they
% have any values, converts those to terms.
%get_rdf_triple_parameters( +Params, -SubjectTerm, -PredicateTerm, -ObjectTerm )
get_rdf_triple_parameters( Params, SubjectTerm, PredicateTerm, ObjectTerm ) :-
    get_http_parameter( Params, subject, SubjectString ),
    get_http_parameter( Params, predicate, PredicateString ),
    get_http_parameter( Params, object, ObjectString ),
    % If the value is empty, leave it as free variable
    normal_string_to_term( SubjectString, SubjectTerm ),
    param_string_to_term( PredicateString, PredicateTerm ),
    normal_string_to_term( ObjectString, ObjectTerm ).


% Converts a http parameter value into a term
%param_string_to_term( +ParamString, -ParamTerm )
param_string_to_term( ParamString, ParamTerm ) :-
    atomic( ParamString ),  % string, but also works for other atomics?
    string_length( ParamString, Length ),
    Length > 0,
    string_to_atom( ParamString, PSAtom ),
    atom_chars( PSAtom, CharList ),
    get_ugly_sub_string_index( CharList, ':', Index ),
    Index > 0,
    parse_http_uri_parameter( ParamString, NameSpaceString, IDString ),
    string_to_atom( NameSpaceString, NSAtom ),
    string_to_atom( IDString, IDAtom ),
    ParamTerm = NSAtom:IDAtom,
    !.

param_string_to_term( ParamString, ParamTerm ) :-
    normal_string_to_term( ParamString, ParamTerm ),
    !.

param_string_to_term( _, _ ).


% Converts a normal string mechanically into an atom (!)
%normal_string_to_term( +ParamString, -ParamTerm )
normal_string_to_term( ParamString, ParamTerm ) :-
    atomic( ParamString ),  % string, but also works for other atomics?
    string_length( ParamString, Length ),
    Length > 0,
    string_to_atom( ParamString, ParamTerm ),
    !.

normal_string_to_term( _, _ ).


% Encodes ampersands correctly.
%encode_ampersands( +String, -EncodedString )

encode_ampersands( String, EncodedString ) :-
    contains_cdata_section( String ),
    EncodedString = String, 
    !.

encode_ampersands( String, EncodedString ) :-
    % numbers: clear case, return as such
    number( String ),
    String = EncodedString.

encode_ampersands( String, EncodedString ) :-
    (
        string( String ),
        WorkString = String;

        atom( String ),
        string_to_atom( WorkString, String );

        sformat( ErrorStr, 'lost_bridge_deflector_shields( \'Bad thing happened, you called encode_ampersands with unknown type\' ): the problematic string=\'~w\'', String ),
        throw( ErrorStr )
    ),
    string_to_list( WorkString, CharList ),
    (
        member( 38, CharList ),  % is the ampersand in the list?
       
        % it was, now let's try to do it the "lazy" way
        string_to_atom( CdataStart, '<![CDATA[' ),
        string_to_atom( CdataEnd, ']]>' ),
        string_concat( CdataStart, WorkString, TempString1 ),
        string_concat( TempString1, CdataEnd, EncodedString );

        EncodedString = WorkString
    ).

encode_gt_lt( Str, Encoded ) :-
    contains_cdata_section( Str ),
    Encoded = Str, !;

    string( Str ),
    encode_gt_lt_1( Str, Encoded ), !;

    atom( str ),
    string_to_atom( WorkString, Str ),
    encode_gt_lt_1( WorkString, Encoded ), !;

    Encoded = Str,
    !.

contains_cdata_section( Input ) :-
    (
        (
            string( Input ),
            Str = Input
        );
        (
            atom( Input ),
            string_to_atom( Str, Input )
        )
    ),
    string( Str ),
    sub_string( Str, _, _, _, '<![CDATA[' ),
    sub_string( Str, _, _, _, ']]>' ),
    !.

    

encode_gt_lt_1( Str, Encoded ) :-
    string_to_list( Str, CharList ),
    (
        member( 60, CharList );  % is the less than '<' in the list?
        member( 62, CharList )   % is the greater than '>' in the list?
    ),
    % it was, now let's try to do it the "lazy" way
    string_to_atom( CdataStart, '<![CDATA[' ),
    string_to_atom( CdataEnd, ']]>' ),
    string_concat( CdataStart, Str, TempString1 ),
    string_concat( TempString1, CdataEnd, Encoded );
    Encoded = Str.

%parse_http_uri_parameter( +URIString, -NameSpaceString, -IDString )
parse_http_uri_parameter( URIString, NameSpaceString, IDString ) :-
    atomic( URIString ),  % string, but also works for other atomics?
    string_to_atom( URIString, URIAtom ),
    atom_chars( URIAtom, CharList ),
    get_ugly_sub_string_index( CharList, '#', Index ),
%    Index >= 0,
%    Index < ?Length,
    sub_string( URIString, 0, Index, _, NameSpaceString ),
    sub_string( URIString, Index, _, 0, IDString ).


%%%%%%%%%%
% Handling the Categories
%


%print_category( +CategoryURI, +LabelingPredicate, +SubStructuringPredicateAtom, +BookmarkPredicateAtom, +Depth )
print_category( CategoryURI, LabelingPredicate, SubStructuringPredicateAtom, BookmarkPredicateAtom, Depth ) :-
    !,   % no back-tracking!
%    category_has_non_empty_subcategories( CategoryURI, SubStructuringPredicateAtom, BookmarkPredicateAtom ),

    % if it is not in the skipping rules
    (
        sewehgrius_category_is_viewable( CategoryURI ),
            writef( '<ogt:Category>\n' ),

        % if we have a set labeling predicate, use it, otherwise use default labeling scheme
        (
            nonvar( LabelingPredicate ),
            call( LabelingPredicate, CategoryURI, Labels );

            sewehgrius_labels( CategoryURI, Labels )
        ),
        http_print_xmllang_list( Labels ),
        http_print_projectionof( CategoryURI ),
        !,  % for some reason this seems to backtrack if there are no annotations

        % print "annotations" (if any)
        get_annotation_list( CategoryURI, Annotations ),
        % add the extra annotation for root categories
        (
            Depth=root,
            % get the root category id count so far FIXXX! Non-thread safe!
            flag( rootcatid, RootCatIdCount, RootCatIdCount+1 ),
            AllAnnotations = [[_, 'rootcatid', literal( RootCatIdCount )]|Annotations];  % or!

            true
        ),
        % call for the actual printing
        http_print_annotations( AllAnnotations );

        % otherwise we do not print the category out at all
        true
    ),

    % get the goals' results into a bag and process it, otherwise just
    % truely pass through

    (
        setof( SubCategoryURI,
            call( SubStructuringPredicateAtom, CategoryURI, SubCategoryURI ),
            SubCategoryBag )
        ->(
            sewehgrius_category_is_viewable( CategoryURI )->(
                writef( '<ogt:subCategories rdf:parseType="Collection">\n' )
            ),
%format('<!--~nnow printing subcategories...~n~nsubcategorybag=~w~n~n-->~n', [SubCategoryBag]),
            print_category_bag( SubCategoryBag, SubStructuringPredicateAtom, BookmarkPredicateAtom ),
            sewehgrius_category_is_viewable( CategoryURI )->(
                writef( '</ogt:subCategories>\n' )
            )
        );

        true
    ),

    % print the bookmarks (or truely pass if none exist and category is viewable)
    (
        sewehgrius_category_is_viewable( CategoryURI ),
        setof( BookmarkURI,
               call( BookmarkPredicateAtom, BookmarkURI, CategoryURI ),
               BookmarkBag )
        ->(
            writef( '<ogt:topicOf rdf:parseType="Collection">\n' ),
            print_bookmark_bag( BookmarkBag ),
            writef( '</ogt:topicOf>\n' )
        );

        true
    ),

    (
        sewehgrius_category_is_viewable( CategoryURI ),
            writef( '</ogt:Category>\n' ),
        flush_output  % we have a large printout to do, flush every now and then
    ).


% either this category or at least one sub-category has bookmarks
%category_has_non_empty_subcategories( +CategoryURI, +SubStructuringPredicateAtom, +BookmarkPredicateAtom )
category_has_non_empty_subcategories( CategoryURI, _, BookmarkPredicateAtom ) :-
    % bagof succeeds if at least one is found  (actually, we wouldn't need the whole bag)
%    bagof( BookmarkURI,
        call( BookmarkPredicateAtom, _, CategoryURI ),  % _ was BookmarkURI
     writef( 'BookmarkSuccess!\n' ).
%        BookmarkBag ),
%    writef( 'BMBag: %w\n', [BookmarkBag] ).

category_has_non_empty_subcategories( CategoryURI, SubStructuringPredicateAtom, BookmarkPredicateAtom ) :-
    bagof( SubCategoryURI,
        call( SubStructuringPredicateAtom, CategoryURI, SubCategoryURI ),
        SubCategoryBag ),
%    writef( 'CATBag: %w\n', [SubCategoryBag] ),
    category_bag_has_non_empty_subcategories( SubCategoryBag, SubStructuringPredicateAtom, BookmarkPredicateAtom ).


%category_bag_has_non_empty_subcategories( +List, SubStructuringPredicateAtom, BookmarkPredicateAtom )
category_bag_has_non_empty_subcategories( [], _, _ ) :-
    writef( 'SUBCATEGORY SYSTEM FAILURE!\n' ),
    fail.


category_bag_has_non_empty_subcategories( [URI|Rest], SubStructuringPredicateAtom, BookmarkPredicateAtom ) :-
    category_has_non_empty_subcategories( URI, SubStructuringPredicateAtom, BookmarkPredicateAtom );
    % if that one didn't have bookmarks, try next one in the list:
    category_bag_has_non_empty_subcategories( Rest, SubStructuringPredicateAtom, BookmarkPredicateAtom ).


%print_category_bag( +URITermList, +SubStructuringPredicateAtom, +InstancePredicateAtom )
print_category_bag( [], _, _ ).

print_category_bag( [URI|List], SubStructuringPredicateAtom, InstancePredicateAtom ) :-
    print_category( URI, _, SubStructuringPredicateAtom, InstancePredicateAtom, branch ),
    print_category_bag( List, SubStructuringPredicateAtom, InstancePredicateAtom ).


%print_bookmark_bag( +List )
print_bookmark_bag( [] ).

print_bookmark_bag( [URI|List] ) :-
    print_w3c_bookmark( URI ),
    print_bookmark_bag( List ).


%http_print_projectionof( +URI )
http_print_projectionof( URI ) :-
    writef( '<ogt:projectionOf rdf:resource="' ),
    writef( URI ),
    writef( '"/>\n' ).


%print_w3c_bookmark( +SubjectURI )
print_w3c_bookmark( SubjectURI ) :-
    writef( '<rdf:Description rdf:about="' ),
    writef( SubjectURI ),
    writef( '"/>\n' ).


%%%%%%%%%%
% Fixing RDF triples' exported values
%


% Brute-force-removes the "literal(' ')"-surrounding
%fix_literal_term( +Literal, -Fixed )
fix_literal_term( Literal, Fixed ) :-
%    atomic( Literal ),
    ComparisonAtom = "literal('",
    string_to_atom( ComparisonString, ComparisonAtom ),
    ComparisonAtom2 = "literal(",
    string_to_atom( ComparisonString2, ComparisonAtom2 ),

    term_to_atom( Literal, LiteralAtom ),
    string_to_atom( NameString, LiteralAtom ), % revert the atom into a string
    (
        (
            sub_string( NameString, 0, 9, _, ComparisonString ),
            sub_string( NameString, 9, _, 2, Fixed )
        );
        (
            sub_string( NameString, 0, 8, _, ComparisonString2 ),
            sub_string( NameString, 8, _, 1, Fixed )
        )
    ).

fix_literal_term( Literal, ResultLiteral ) :-
    atomic( Literal ),
    ResultLiteral = Literal.

fix_literal_term( _, _ ) :- fail.


% Returns the string in the URL after the last slash
%get_predicate_name_portion_of_url( +Url, -PredicateName )
get_predicate_name_portion_of_url( Url, PredicateName ) :-
    atom_chars( Url, Char_list ),
    get_ugly_sub_string_index( Char_list, '/', Index ),
    string_length( Url, Length ),
    Index < Length,
    sub_string( Url, Index, _, 0, PredicateNameTemp ),
    get_predicate_name_portion_of_url( PredicateNameTemp, PredicateName ),
    !.

get_predicate_name_portion_of_url( Url, PredicateName ) :-
    atom_chars( Url, Char_list ),
    get_ugly_sub_string_index( Char_list, '/', Index ),
    string_length( Url, Length ),
    Index < Length,
    sub_string( Url, Index, _, 0, PredicateName ).


%get_ugly_sub_string_index( +List, +Char, -Index ) :-
get_ugly_sub_string_index( List, Char, Index ) :-
    List = [FirstChar|Temp_list],
    not( FirstChar = Char ),
    get_ugly_sub_string_index( Temp_list, Char, LastIndex ),
    Index is LastIndex + 1,
    !.

get_ugly_sub_string_index( [], _, _ ) :-
    fail,
    !.

get_ugly_sub_string_index( _, _, LastIndex ) :-
    LastIndex is 1,
    !.


foreach( Goal, Action ) :-
    forall( setof( _, Goal, _ ), Action ).

foreach( Sort, Goal, Action ) :-
    setof( Sort, Goal, List ),
    checklist( Action, List ).


normalize_uri( NS:ID, URI ) :-
        !,
        concat_atom( [NS,ID], URI ).

normalize_uri( IN, OUT ) :-
        OUT=IN.

% Merges all sub-lists in a relation weight recommendation score
% calculation hairy-as-in-hedgehog List and returns the result in
% SuperList (duplicates are not removed). So in essence, this is
% sort of a "tree-merge", which merges to "all directions below and
% to the left and right".
merge_entire_listtree( [], [] ).

merge_entire_listtree( [SubList|Tail], SuperList ) :-
    merge_entire_listtree( Tail, RemainderSuperList ),
    (
        % empty sub-list, skip
        SubList = [],
        SuperList = RemainderSuperList;

        % non-empty sub-list/item, add.
        % if it is anything containing four things,
        % add it as such (it is a result as its own)
        SubList = [_, _, _, _],
        SuperList = [SubList|RemainderSuperList];

        % nope, it appears to be some deeper structure --
        % ad traversam, mon ami!
        merge_entire_listtree( SubList, SubSubList ),
        merge( SubSubList, RemainderSuperList, SuperList )
    ).

merge_entire_listtree( A, B ) :-
    writelog( 'Merge_entire_listtree got: ~w and ~w~n', [A, B] ),
    throw( http_status( 500, 'Merge_entire_listtree could not handle the type!' ) ).



%%%%%%%%%%
% Predicates for quick generic functionality testing
%


test_categories :-
    string_to_atom( '/predicate/get_categories', A ),
    reply_core( A, []).

test_categories2 :-
    string_to_atom( '/predicate/get_categories2', A ),
    reply_core( A, [uri='http://foo/a#b']).

test_annotations :-
    string_to_atom( '/predicate/get_annotations', A ),
    reply_core( A, [uri='http://www.cs.helsinki.fi/group/seco/ns/2004/03/18-esinekortti#EKM1'] ),
    print_out_http_ok.

test_relations :-
    test_relations( 'http://www.cs.helsinki.fi/group/seco/ns/2004/03/18-esinekortti#EKM1' ).

test_relations( SubjectURI ) :-
    Params = [subject=SubjectURI],
    string_to_atom( '/predicate/get_relations', A ),
%    reply_core( A, [subject='http://www.cs.helsinki.fi/group/seco/ns/2004/03/18-esinekortti#kortti103',
%        object='http://www.cs.helsinki.fi/group/seco/ns/2004/03/18-esinekortti#kortti200' ] ),
    reply_core( A, Params ),
true.
%    print_out_http_ok.  %everything went ok, so this will only be executed when successful





%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% My simple cache for uripairs.
%
% The usage of cache may be problematic since the amount of cached uripairs can be as high as C^2, where 
% C is the amount of itemcards.

add_to_uripair_cache(SubjectURI, ObjectURI, Score) :-
	writelog( "adding to uripair cache: SubjectURI: ~w; ObjectURI: ~w; Score: ~w", [SubjectURI, ObjectURI, Score]),
	assert(cache_uripair(SubjectURI, ObjectURI, Score)).

get_from_uripair_cache(SubjectURI, ObjectURI, Score) :-
	nonvar(SubjectURI),
	nonvar(ObjectURI),
	catch(
 	     cache_uripair(SubjectURI, ObjectURI, Score),
	     _,
	     fail),
%	!,
%	writelog( "Fetched from uripair cache: [~w, ~w, ~w]~n", [SubjectURI, ObjectURI, Score] ),
	!.

% distancesquared(+X1,+Y1,+X2,+Y2,-D)
distancesquared(X1,Y1,X2,Y2,D) :-
        D is (X1-X2)*(X1-X2)+(Y1-Y2)*(Y1-Y2).

locationinfo_muinainen(C,CX,CY) :-
        muinainen(C),
        rdf( C, 'http://www.cs.helsinki.fi/group/seco/ns/2004/03/18-muinainen#':'WGS84_X-koordinaatti', literal(CXAtom)),
        atom_to_integer( CXAtom, CX),
	    rdf( C, 'http://www.cs.helsinki.fi/group/seco/ns/2004/03/18-muinainen#':'WGS84_Y-koordinaatti', literal(CYAtom) ),
	    atom_to_integer( CYAtom, CY).
	
% locationinfo(-C,-CX,-CY)
locationinfo(C,CX,CY) :-
        paikka(C),
        rdf( C, 'http://www.cs.helsinki.fi/group/seco/ns/2004/03/18-paikat#':'koordinaatti', K),
        rdf( K, 'http://www.cs.helsinki.fi/group/seco/ns/2004/03/18-paikat#':'WGS84_X-koordinaatti', literal(CXAtom)),
        atom_to_integer( CXAtom, CX),
	    rdf( K, 'http://www.cs.helsinki.fi/group/seco/ns/2004/03/18-paikat#':'WGS84_Y-koordinaatti', literal(CYAtom) ),
	    atom_to_integer( CYAtom, CY).
	    
% locationnearest(-D, C, X, Y)
locationnearest(D,C, X, Y) :-
		nonvar(C),
		nonvar(X),
		nonvar(Y),
		paikka(C),
		findall(D,(locationinfo(C,CX,CY),distancesquared(X,Y,CX,CY,D)),ResultList),
		sort(ResultList,SortedResultList),
	    nth0(0,SortedResultList,D).

% nearest(+X,+Y,-C)
nearest(X,Y,SortedResultList) :-
    nonvar(X),
    nonvar(Y),
    findall([D,C],(paikka(C),locationnearest(D,C,X,Y)),ResultList),
    sort(ResultList,SortedResultList).

nearest_muinainen(X,Y,SortedResultList) :-
    nonvar(X),
    nonvar(Y),
    findall([D,C],(muinainen(C),locationinfo_muinainen(C,CX,CY),distancesquared(X,Y,CX,CY,D)),ResultList),
    sort(ResultList,SortedResultList).

% print_location(+C,+D)
print_location(D,C) :-
	nonvar(C),
	nonvar(D),
	E is sqrt(D),
	rdf(C,'http://www.cs.helsinki.fi/group/seco/ns/2004/03/18-paikat#':'nimi',literal(L)),
	http_print( '<swhg:location rdf:resource="'),
	http_print(C),
	http_print('" swhg:distance="'),
	http_print(E),
	http_print('" rdfs:label="'),
	http_print(L),
	http_print('"/>\n').

print_location_muinainen(D,C) :-
	nonvar(C),
	nonvar(D),
	E is sqrt(D),
	rdf(C,'http://www.cs.helsinki.fi/group/seco/ns/2004/03/18-muinainen#':'www_nimi',literal(L)),
	http_print( '<swhg:location rdf:resource="'),
	http_print(C),
	http_print('" swhg:distance="'),
	http_print(E),
	http_print('" rdfs:label="'),
	http_print(L),
	http_print('"/>\n').

get_nearest( Params ) :-

%% Print to log

    writelog( "calling: get_nearest with ~w", [Params]),
	get_http_parameter( Params, x, XString ),
	get_http_parameter( Params, y, YString ),
	atom_to_integer( XString, X ),
	atom_to_integer( YString, Y ),
	nearest(X,Y,SRL),
	
    print_out_http_ok,
    http_print_rdf_doctype,
    http_begin_rdf_document,

    http_print( '<swhg:NearestLocation>\n' ),
    http_print( '<swhg:nearest rdf:parseType="Collection">\n' ),
	nth0(0,SRL,[D1,C1]),
	print_location(D1,C1),
	nth0(1,SRL,[D2,C2]),
	print_location(D2,C2),
	nth0(2,SRL,[D3,C3]),
	print_location(D3,C3),
	nth0(3,SRL,[D4,C4]),
	print_location(D4,C4),
	nth0(4,SRL,[D5,C5]),
	print_location(D5,C5),
    http_print( '</swhg:nearest>\n' ),
    http_print( '</swhg:NearestLocation>\n' ),

    % the multi-URI loop should end here

    http_end_rdf_document,
    !.

get_nearest_muinainen( Params ) :-

%% Print to log

    writelog( "calling: get_nearest_muinainen with ~w", [Params]),
	get_http_parameter( Params, x, XString ),
	get_http_parameter( Params, y, YString ),
	atom_to_integer( XString, X ),
	atom_to_integer( YString, Y ),
	nearest_muinainen(X,Y,SRL),
	
    print_out_http_ok,
    http_print_rdf_doctype,
    http_begin_rdf_document,

    http_print( '<swhg:NearestMuinainen>\n' ),
    http_print( '<swhg:nearest rdf:parseType="Collection">\n' ),
	nth0(0,SRL,[D1,C1]),
	print_location_muinainen(D1,C1),
	nth0(1,SRL,[D2,C2]),
	print_location_muinainen(D2,C2),
	nth0(2,SRL,[D3,C3]),
	print_location_muinainen(D3,C3),
	nth0(3,SRL,[D4,C4]),
	print_location_muinainen(D4,C4),
	nth0(4,SRL,[D5,C5]),
	print_location_muinainen(D5,C5),
	nth0(5,SRL,[D6,C6]),
	print_location_muinainen(D6,C6),
	nth0(6,SRL,[D7,C7]),
	print_location_muinainen(D7,C7),
	nth0(7,SRL,[D8,C8]),
	print_location_muinainen(D8,C8),
	nth0(8,SRL,[D9,C9]),
	print_location_muinainen(D9,C9),
	nth0(9,SRL,[D10,C10]),
	print_location_muinainen(D10,C10),
	nth0(10,SRL,[D11,C11]),
	print_location_muinainen(D11,C11),
	nth0(11,SRL,[D12,C12]),
	print_location_muinainen(D12,C12),
    http_print( '</swhg:nearest>\n' ),
    http_print( '</swhg:NearestMuinainen>\n' ),

    % the multi-URI loop should end here

    http_end_rdf_document,
    !.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% My help predicates
%


% Print to log
writelog(FormatString, Values) :-
	sformat( String, FormatString, Values ),
	writelog( String ).

writelog(String) :-
	get_config_use_logging,
	thread_send_message( httpdlogger, String ).
